#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <malloc.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

enum
{
	MODE_GENERIC,
	MODE_GII,
	MODE_MII,
	MODE_CII
};

int mode;
int rl_initted;
static char hist_file[PATH_MAX];

char *passprompt;
char *recvprompt;
char *prompt;
char guessed_prompt[256];
char guessed_oprompt[256];

void finish(void) __attribute__((noreturn));


int send_bytes(int fd, char *buf, int len)
{
	return write(fd, buf, len);
}


int expect(int infd, char *str, FILE *ofp)
{
	int k, n;
	int pattern_len = strlen(str);
	char buf[4096];

	n = 0;
	while ((k = read(infd, buf+n, sizeof(buf)-n)) > 0) {
		n += k;
		if (n >= pattern_len &&
		    memcmp(buf+n-pattern_len, str, pattern_len) == 0) {
			if (n != pattern_len)
				fwrite(buf, n-pattern_len, 1, ofp);
			fflush(ofp);
			return 0;
		}
		if (buf[n-1] == '\n') {
			fwrite(buf, n, 1, ofp);
			n = 0;
		} else if (n == sizeof(buf)) {
			fwrite(buf, sizeof(buf)/2, 1, ofp);
			memcpy(buf, buf+sizeof(buf)/2, sizeof(buf)/2);
			n = sizeof(buf)/2;
		}
	}
	return -1;
}

int slow_expect(int infd, char *str, FILE *ofp, int guess)
{
	int n;
	int i;
	int pattern_len = strlen(str);
	unsigned char buf[4096];

	while ((n = read(infd, buf, sizeof(buf))) > 0) {
		for (i=0; i<n; i++) {
			if (buf[i] == 255) {
				while (i+3 > n) {
					int k = read(infd, buf+n, 1);
					if (k <= 0)
						finish();
					n += k;
				}
				if (buf[i+1] >= 251 || buf[i+1] <= 254) {
					buf[i+1] = 254;
					send_bytes(infd, buf+i, 3);
				}
				memmove(buf+i, buf+i+3, n-i-3);
				i -= 3;
				n -= 3;
			}
		}
		if (n >= pattern_len &&
		    memcmp(buf+n-pattern_len, str, pattern_len) == 0) {
			if (guess) {
				char *gp;
				buf[n] = 0;
				gp = strrchr(buf, '\n');
				if (gp == NULL)
					gp = buf;
				else
					gp++;
				strcpy(guessed_prompt, gp);
				snprintf(guessed_oprompt, sizeof(guessed_oprompt), "%.3s/%s", prompt, guessed_prompt);
				prompt = guessed_oprompt;
				pattern_len = strlen(gp);
			}
			if (n != pattern_len)
				fwrite(buf, n-pattern_len, 1, ofp);
			fflush(ofp);
			return 0;
		}
		fwrite(buf, n, 1, ofp);
	}
	return -1;
}

static char rl_fake_buf[256];
int rl_fake_point;

void rl_dirty_trick(void)
{
	extern char *rl_line_buffer;
	extern int rl_point, rl_end;
	extern void (*rl_startup_hook)();

	rl_startup_hook = NULL;

	strcpy(rl_line_buffer, rl_fake_buf);
	rl_point = rl_fake_point;
	rl_end = strlen(rl_line_buffer);
}

int
main(int argc, char **argv)
{
	struct sockaddr_in dst;
	int sock;
	FILE *fpout = stdout;
	char *pass;
	struct hostent *he;
	char *portname = NULL;
	char *progname = basename(argv[1]);
	char *hist_basename;
	char *help_mode=NULL;
	char expected[256];
	int eat = 0;

	if (argc<2 || argc>3) {
		fprintf(stderr, "Usage: %s <destination> <port>\n", progname);
		exit(1);
	}
	he = gethostbyname(argv[1]);
	if (he == NULL) {
		fprintf(stderr, "gethostbyname: resolution %s failed\n", argv[1]);
		exit(1);
	}
	memset(&dst, 0, sizeof(dst));
	dst.sin_family = AF_INET;
	memcpy(&dst.sin_addr, he->h_addr, 4);
	if (argc > 2)
		portname = argv[2];

	mode = MODE_GENERIC;

	if (portname) {
		if (strcmp(portname, "gii") == 0)
			mode = MODE_GII;
		else if (strcmp(portname, "mrt") == 0)
			mode = MODE_MII;
		else if (strcmp(portname, "telnet") == 0)
			mode = MODE_CII;
	}

	if (mode == MODE_GENERIC) {
		if (strcmp(progname, "gii") == 0) {
			mode = MODE_GII;
		} else if (strcmp(progname, "mii") == 0) {
			mode = MODE_MII;
		} else if (strcmp(progname, "cii") == 0) {
			mode = MODE_CII;
		}
	}

	switch (mode) {
	case MODE_GENERIC:
		hist_basename = ".";
		passprompt = NULL;
		recvprompt = "> ";
		prompt = "ii> ";
		break;
	case MODE_GII:
		hist_basename = ".gii_history";
		passprompt = "Password? ";
		recvprompt = "> ";
		prompt = "gii> ";
		if (!portname)
			portname = "gii";
		break;
	case MODE_MII:
		hist_basename = ".mii_history";
		passprompt = "password> ";
		recvprompt = "MRT> ";
		prompt = "mii> ";
		if (!portname)
			portname = "mrt";
		break;
	case MODE_CII:
		hist_basename = ".cii_history";
		passprompt = "Password: ";
		recvprompt = ">";
		prompt = "cii> ";
		if (!portname)
			portname = "telnet";
		break;
	}

	strcpy(guessed_prompt, recvprompt);
	recvprompt = guessed_prompt;

	if (portname == NULL) {
		fprintf(stderr, "Hey, what service did you request?\n");
		exit(1);
	}

	if (portname[0] > '0' && portname[1] <= '9') {
		dst.sin_port = htons(atoi(portname));
	} else {
		struct servent *se = getservbyname(portname, "tcp");
		if (se == NULL) {
			fprintf(stderr, "getservbyname: resolution %s failed\n", portname);
			exit(1);
		}
		dst.sin_port = se->s_port;
	}

	strcpy(hist_file, hist_basename);
	if (getenv("HOME") == NULL || access(hist_file, R_OK)) {
		if (getenv("HOME"))
			snprintf(hist_file, sizeof(hist_file),
				 "%s/%s", getenv("HOME"), hist_basename);
	}

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock<0) {
		perror("socket");
		exit(1);
	}
	if (connect(sock, (struct sockaddr*)&dst, sizeof(dst)) < 0) {
		perror("connect");
		exit(1);
	}

	if (passprompt) {
		if (slow_expect(sock, passprompt, fpout, 0) < 0)
			finish();
		pass = getpass("Password: ");
		if (pass == NULL)
			finish();
		send_bytes(sock, pass, strlen(pass));
		send_bytes(sock, "\r\n", 2);
	}

	if (slow_expect(sock, recvprompt, fpout, mode != MODE_MII) < 0)
		finish();

	if (mode == MODE_CII) {
		send_bytes(sock, "terminal length 0\r\n", sizeof("terminal length 0\r\n")-1);
		if (expect(sock, recvprompt, fpout) < 0)
			finish();
	}
	strcpy(expected, recvprompt);

	do {
		char *p, *p1;

		strcpy(expected, recvprompt);
		if (fpout != stdout) {
			fclose(fpout);
			fpout = stdout;
		}
		if (help_mode) {
			while (eat>0) {
				send_bytes(sock, "\b", 1);
				eat--;
			}
			help_mode = NULL;
		}
		if (!rl_initted) {
			read_history(hist_file);
			rl_initted = 1;
		}
		p = readline(prompt);
		if (p==NULL)
			finish();
		if (*p)
			add_history(p);
		p1 = strchr(p, '>');
		if (!p1)
			p1 = strchr(p, '|');
		if (p1) {
			int pipe=0, append=0;
			if (*p1 == '>') {
				*p1 = 0;
				p1++;
				if (*p1 == '>') {
					p1++;
					append=1;
				}
			} else {
				pipe=1;
				*p1 = 0;
				p1++;
			}
			while (*p1 == ' ')
				p1++;
			if (pipe) {
				signal(SIGPIPE, SIG_IGN);
				fpout = popen(p1, "w");
			} else {
				fpout = fopen(p1, append?"a":"w");
			}
			if (fpout==NULL) {
				fpout = stdout;
				fprintf(stderr, "Cannot redirect to %s\n", p1);
				continue;
			}
		}
		help_mode = NULL;
		if (p[0]) {
			if (mode == MODE_MII || mode == MODE_CII)
				help_mode=strchr(p, '?');
			if (help_mode) {
				extern void (*rl_startup_hook)();
				
				send_bytes(sock, p, help_mode-p+1);
				snprintf(expected, sizeof(expected),
					 "%s%*.*s", recvprompt,
					 help_mode-p,
					 help_mode-p,
					 p);
				rl_startup_hook = rl_dirty_trick;
				rl_fake_point = help_mode-p;
				eat = rl_fake_point;
				memcpy(rl_fake_buf, p, rl_fake_point);
				strncpy(rl_fake_buf+rl_fake_point, help_mode+1, sizeof(rl_fake_buf)-rl_fake_point);
			} else {
				send_bytes(sock, p, strlen(p));
				send_bytes(sock, "\r\n", 2);
			}
		} else
			send_bytes(sock, "\r\n", 2);
		free(p);
	} while (expect(sock, expected, fpout) == 0);

	finish();
}

void finish()
{
	if (rl_initted)
		write_history(hist_file);
	exit(0);
}
