/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

#define GIIMAXCMDLEN	BUFSIZ
#define GIIMAXLINELEN	GIIMAXCMDLEN
#define GIINBMAXTOKENS	30
#define GIITOKENLEN	80
#define GII_MAXTRIES	3

/* Some global flags
 */
#define GIIF_ON		1		/* GII is ON */

/* The structure to store the command names and the associated command
 * to execute.
 */
typedef struct _giimenu_cmd_t {
	char *cmd_name;			/* Name of the command */
	struct _giimenu_cmd_t *cmd_nextCmdTable;	/* The subcommands */
	int (*cmd_lasttok)();		/* What to execute if there is no more
					 * token to read but the command
					 * need a subcommand (error) */
	int (*cmd_exec)();		/* What to execute if there is no
					 * subcommands */
	char *cmd_helpMsg;		/* An explanation of the command */
} giimenu_cmd_t;


/* Internal state of a GII session
 */
typedef struct _gii_ctl_t {
	struct _gii_ctl_t *g_prev, *g_next;	/* Linked list */
	int g_state;			/* The state. C.f. GIIS_XXX */
	task *g_task;			/* Gated task for that session */
	giimenu_cmd_t *g_cmdmenu;	/* Current command menu */
	giimenu_cmd_t *g_cmd;		/* Current command in that menu */
	char g_cmdStr[GIIMAXCMDLEN];	/* The current parsed command */
	char g_buff[2*GIIMAXLINELEN];	/* Line buffer */
	int g_bufflen;			/* what is the buffer */
	int g_buffLeft;			/* what is left free */
	int g_iden_tries;		/* Number of identication tries */
	rtwalk_t *g_walk;		/* Used when walking the routing tabl */
	task_job *g_job;		/* Any pending job */
#ifdef PROTO_OSPF2
	ospf_ptwalk_t *g_ospfwalk;	/* Used to walk the OSPF tree */
#endif
} gii_ctl_t;

/* State of a session
 */
#define GIIS_OPEN	0		/* Before authentication */
#define GIIS_SESSION	1		/* After auth. */
#define GIIS_JOB	2		/* A job is running... */
#define GII_LASTSTATE	2		/* The last one */

/* Default user name when authenticating
 */
#define GII_USER	"gii"
#define GIIPROMPT	"GateD-%s> "
#define GII_PORT	616

#define GIISTATESTR(state)	(((state) > GII_LASTSTATE)? "???": \
		gii_statenames[(state)])

/* Codes when printing messages
 */
#define GW_NONE		0		/* No code */
#define GW_ERR		5		/* An error message */
#define GW_INFO		1		/* An information message */

/* Strip any whierd character at the begining of a string
 */
#define STRIP(ln)	{ \
	char *c = ln + strlen(ln) - 1; \
	while(*(ln) == '\n' || *(ln) == '\r') (ln)++; \
	while(*c == '\n' || *c == '\r') *(c--) = '\0'; \
}
#define GII_WRITE(u)	do { if (gii_write u) return(1); } while(0)
#define GII_NOAVAILABLE	gii_write(gii_ctl, GW_ERR, "function not avaliable")

void		gii_init __P((void));
void		gii_var_init __P((void));
void		gii_accept __P((task *));
int		gii_iden __P((gii_ctl_t *));
void		gii_recv __P((task *));
int		gii_process __P((gii_ctl_t *, char *));
int		gii_write __P((gii_ctl_t *, int, char *, ...));
void		gii_terminate __P((task *));
void		gii_cleanup __P((void));
void		gii_dump __P((task *, FILE *));
int 		gii_parse_cmd __P((gii_ctl_t *, char *));
int 		gii_cmd_error __P((gii_ctl_t *));
int		split __P((char *, char[][]));
giimenu_cmd_t *	cmd_find __P((giimenu_cmd_t [], char *));
int		gii_quit __P((gii_ctl_t *));
int		gii_showversion __P((gii_ctl_t *));
int		gii_showkernel __P((gii_ctl_t *));
int		gii_help __P((gii_ctl_t *));
int		gii_prompt __P((gii_ctl_t *));
void		telnet_strip __P((char *, int *));
int		telnet_echooff __P((gii_ctl_t *));
int		telnet_echoon __P((gii_ctl_t *));
char *		trace_state_all __P((flag_t));
int		gii_showif __P((gii_ctl_t *, char [][], int));
int		gii_showallif __P((gii_ctl_t *));
int		gii_showmem __P((gii_ctl_t *));
int		gii_showrtipall __P((gii_ctl_t *));
int		gii_showrtip __P((gii_ctl_t *, char [][], int));
int		gii_showalltask __P((gii_ctl_t *));
int		gii_showalltimer __P((gii_ctl_t *));
int		gii_showipup __P((gii_ctl_t *, char [][], int));
int		gii_showipdown __P((gii_ctl_t *, char [][], int));
void		gii_job_walk __P((task_job *));
int		gii_dvmrp_mfcall __P((gii_ctl_t *));
int		gii_dvmrp_mfc __P((gii_ctl_t *, char [][], int));
int		gii_dvmrp_targets __P((gii_ctl_t *));
int		gii_ospf_global __P((gii_ctl_t *));
int		gii_ospf_allintf __P((gii_ctl_t *));
int		gii_ospf_intf __P((gii_ctl_t *, char [][], int));
int		gii_ospf_allarea __P((gii_ctl_t *));
int		gii_ospf_area __P((gii_ctl_t *, char [][], int));
int		gii_ospf_showdb __P((gii_ctl_t *));
void		gii_job_ospftree __P((task_job *));

