/*
 * rt_router.c
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 */

/* Saves lots of memory (and memory bandwidth) on BGP routers
   by sharing nexthop addresses between routes.
 */

#define	INCLUDE_RT_VAR

#include "include.h"
#ifdef	PROTO_INET
#include "inet.h"
#endif				/* PROTO_INET */
#ifdef	PROTO_ISO
#include "iso.h"
#endif				/* PROTO_ISO */
#include "krt.h"

static block_t rtr_block_index;

struct rtr_entry
{
	struct rtr_entry *rtr_next;
	unsigned 	rtr_magic;
	int		rtr_users;
	char		rtr_data[0];
};

static struct rtr_entry  *rtr_ht[64];
int tot_rtr;

sockaddr_un * rtr_sockdup(sockaddr_un *r)
{
	unsigned h = r->in.gin_addr.s_addr & (64-1);
	struct rtr_entry * re;

	++tot_rtr;

	for (re = rtr_ht[h]; re; re = re->rtr_next) {
		sockaddr_un *s = (sockaddr_un*)re->rtr_data;

		if (s->in.gin_addr.s_addr == r->in.gin_addr.s_addr) {
			re->rtr_users++;
			return (sockaddr_un*)re->rtr_data;
		}
	}

	if (!rtr_block_index)
		rtr_block_index = task_block_init(sizeof (struct rtr_entry) + 8, "routers");

	re = task_block_alloc(rtr_block_index);
	memcpy(re->rtr_data, r, 8);
	re->rtr_users = 1;
	re->rtr_magic = 0x5A231966;
	re->rtr_next = rtr_ht[h];
	rtr_ht[h] = re;
	return (sockaddr_un*)re->rtr_data;
}

void rtr_sockfree(sockaddr_un *r)
{
	struct rtr_entry * re = ((struct rtr_entry*)r) - 1;

	--tot_rtr;

	if (re->rtr_magic != 0x5A231966) {
		syslog(LOG_ERR, "free not router %08lx : %d\n", ntohl(r->in.gin_addr.s_addr), tot_rtr);
		return;
	}
	if (--re->rtr_users < 0) {
		syslog(LOG_ERR, "router underflow %08lx : %d\n", ntohl(r->in.gin_addr.s_addr), tot_rtr);
		return;
	}
}
